
/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <wba/wba.h>
#include <bld/scope.h>
#include <bld/interp.h>

/***********************************************************************
class BLD_Scope
***********************************************************************/
BLD_Scope::BLD_Scope(void) : variables(GPL_String::HashFunction, 16)
{
}

BLD_Scope::BLD_Scope(const GPL_String & n) : name(n), variables(GPL_String::HashFunction, 16)
{
}

BLD_Scope::~BLD_Scope(void)
{
}

void BLD_Scope::GetValues(GPL_Map<GPL_String, BLD_Value> & values) const
{
	GPL_MapIterator<GPL_String, BLD_Value> iterator;
	iterator.Reset((GPL_Map<GPL_String, BLD_Value> *) &variables);
	while (iterator.Next())
	{
		values.Replace(iterator.Key(), iterator.Value());
	}
}

GPL_Boolean BLD_Scope::LookUp(const GPL_String & string,
								BLD_Value & value) const
{
	return variables.LookUp(string, value);
}

GPL_Boolean BLD_Scope::Insert(const GPL_String & string,
								const BLD_Value & value)
{
	variables.Find(string) = value;
	return TRUE;
}

GPL_Boolean BLD_Scope::AddRule(const GPL_String & string,
								const BLD_RuleType type,
								const OpArray & ops)
{
	// Check for meta-characters after any variables have been replaced
	// with arbitrary (but current) values.  If the variables were left
	// in, they would cause the rather simple algorithm of
	// BLD_RuleBase::IsPattern() to incorrectly identify as patterns
	// strings which only contained variables.
	register const GPL_Boolean isPattern =
		BLD_RuleBase::IsPattern(BLD_CONTEXT.Evaluate(string));

	// Check for at least one variable.
	register const GPL_Boolean isVariable =
		BLD_RuleBase::IsVariable(string);

	// Determine which type of rule to use.
	GPL_Pointer<BLD_RuleBase> rule;
	if (!isPattern)
	{
		if (!isVariable)
		{
			rule = (BLD_RuleBase *) new BLD_Rule(string, type, ops);
		}
		else
		{
			rule = (BLD_RuleBase *) new BLD_VariableRule(string, type, ops);
		}
	}
	else
	{
		// Escape any dots.  Yeah, this prevents the use of dots
		// in pattern expressions, but it *greatly* simplifies the
		// majority of patterns since dots are almost universally
		// used to separate base names from extensions.  Besides,
		// most cases where someone would use '.', they are being
		// lazy and imprecise.  They most likely want to restrict
		// the character range to at least printable characters.
		GPL_String pattern = string.Escape(".");

		// Anchor the pattern if it is not already anchored.
		static GPL_RegularExpression anchoredExpression("\\$$");
		if (!anchoredExpression.Matches(pattern))
		{
			pattern.Append('$');
		}

		if (!isVariable)
		{
			rule = (BLD_RuleBase *) new BLD_PatternRule(pattern, type, ops);
		}
		else
		{
			rule = (BLD_RuleBase *) new BLD_VariablePatternRule(pattern, type, ops);
		}
	}

	// Add the rule to the list.
	rules.Append(rule);
	return TRUE;
}

GPL_Boolean BLD_Scope::CreateTarget(BLD_Target & target)
{
	// Check the rules.
	GPL_Boolean hasBeenFound = FALSE;
	GPL_Array<GPL_String> expressions(1);
	register const unsigned long n = rules.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		if (rules[i]->Matches(target.GetName(), expressions))
		{
			hasBeenFound = TRUE;
			target.AddRule(rules[i], expressions);
		}
	}
	return hasBeenFound;
}

GPL_String BLD_Scope::ToString(void) const
{
	GPL_String image;
	image.Printf("(%p) %s:", this, name.GetData());
	GPL_String value;
	GPL_MapIterator<GPL_String, BLD_Value> iterator;
	iterator.Reset((GPL_Map<GPL_String, BLD_Value> *) &variables);
	while (iterator.Next())
	{
		value = iterator.Key();
		value.Append('=');
		value.Append(iterator.Value().GetString());
		image.Join("\n", value);
	}
	return image;
}
