/******************************************************************************

Coral Tree spa/error.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <spa/spa.h>

WBA_INIT_DEFINITION(SPA);

OSD_Output spaGlobalOutput;
#if OSD_OSTYPE==OSD_UNIX
unsigned long spaGlobalOutputInitialFlags = OSD_OUT_STDERR;
#elif OSD_OSTYPE==OSD_WIN32
unsigned long spaGlobalOutputInitialFlags = OSD_OUT_STDERR;
#endif

long SPA_GLOBAL_INDENT = 0;

long SPA_GLOBAL_VERBOSITY = OSD_VERBOSITY;

long SPA_GLOBAL_INITIALIZED = 0;

long SPA_GLOBAL_STARTTIME = 0;
long SPA_GLOBAL_LASTTIME;

void spaPrintInitialize()
{
	if(!SPA_GLOBAL_INITIALIZED)
	{
		osdOutputOpen(&spaGlobalOutput);
		osdOutputSetFlags(&spaGlobalOutput,spaGlobalOutputInitialFlags);
		SPA_GLOBAL_INITIALIZED = 1;
	}
}


/******************************************************************************
void	spaPrintWithLocation(long source,long type,char *message,char *file,long line)

	print given message, file name, and line number

	this function can be adjusted to react to messages of
		specific sources and types

	TODO: vararg this like printf

Keywords: debug, c, blackbox_system, system
******************************************************************************/
void spaPrintWithLocation(long source,long type,char *message,char *file,long line,long indent)
{
	int i;
	static char sourcename[4][20]=  { 	"SOFTWARE",
										"SYSTEM",
										"USER",
										"GENERAL" };
	static char tname[5][20]=	{ 	"LOG", 
										"WARNING", 
										"ERROR", 
										"FATAL"};

	static char title[SPA_MAX_PRINT_TITLE_LEN];
	static char string[SPA_MAX_PRINT_STRING_LEN];
	static char leader[SPA_MAX_PRINT_STRING_LEN];

	spaPrintInitialize();

	for(i = 0; i < indent;i++)
	{
		leader[i] = ' ';
	}
	leader[i] = 0;

	if(SPA_GLOBAL_VERBOSITY == OSD_DEBUG)
	{
		sprintf(string,"%s%s FILE: %s LINE: %d ERRNUM: %d",
			leader, message, file, (int)(line), OSD_ERRNO);
	}
#if 0
	else if(SPA_GLOBAL_VERBOSITY == OSD_BETA)
	{
		if(type > 1)
		{
			sprintf(string,"%s%s FILE: %s LINE: %d ERRNUM: %d",
				leader, message, file, line, OSD_ERRNO);
		}
		else
		{
			sprintf(string,"%s%s",leader,message);
		}
		/* SHUT UP COMPILER */
		file = file;
		line = line;
	}
#endif
	else if(SPA_GLOBAL_VERBOSITY == OSD_NORMAL)
	{
		if(type > 1)
		{
			sprintf(string,"%s%s FILE: %s LINE: %d ERRNUM: %d",
				leader, message, file, (int)(line), OSD_ERRNO);
		}
		else
		{
			sprintf(string,"%s%s",leader,message);
		}
		/* SHUT UP COMPILER */
		file = file;
		line = line;
	}
	else if(SPA_GLOBAL_VERBOSITY == OSD_SILENT)
	{
		/* SHUT UP COMPILER */
		file = file;
		line = line;
	}

	if(SPA_GLOBAL_VERBOSITY != OSD_SILENT)
	{
		sprintf(title,"SPA %s %s",sourcename[source],tname[type]);
		osdOutputPrintLine(&spaGlobalOutput,title,string);
	}

	/* EXTRA BEHAVIOR */
	switch(type)
	{
		case SPA_LOG:
			/* NOOP */
			break;
		case SPA_WARN:
			/* NOOP */
			break;
		case SPA_ERROR:
			exit(-1);
			break;
		case SPA_FATAL:
			exit(-1);
			break;
		default:
			osdOutputPrintLine(&spaGlobalOutput,"PRINT ERROR","invalid message type");
			exit(-2);
			break;
	}

}

void spaPrintf(char *format, ...)
{
	va_list args;

	spaPrintInitialize();
	va_start(args, format);
	osdVPrintf(&spaGlobalOutput, "SPA", format, args);
	va_end(args);
}


