/******************************************************************************

Coral Tree spa/obfenc.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <spa/spa.h>

static const SPA_Randomizer randomizer(osdTimer());

static void spaPad(GPL_Block & data)
{
	const GPL_Block d = data;
	register const unsigned long n = d.GetSize();
	data.Resize(n * 2);
	register unsigned long j = 0;
	register unsigned long i = 0;
	while (n > i)
	{
		data[j++] = (GPL_Byte) (0xFF & randomizer.RandomNumber());
		data[j++] = d[i++];
	}
}

static void spaUnpad(GPL_Block & data)
{
	const GPL_Block d = data;
	register const unsigned long n = d.GetSize();
	data.Resize((n + 1) / 2);
	register unsigned long i = 0;
	register unsigned long j = 0;
	while (n > i)
	{
		data[j++] = d[++i];
		i++;
	}
}

#if 0
static void printBytes(const GPL_Block & data)
{
	register const unsigned long n = data.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		fprintf(stderr, "%02X", (int) (0xff & (data[i])));
	}
	fprintf(stderr, "\n");
	fflush(stderr);
}
#endif

#if 0
static void printBits(const GPL_Block & data)
{
	register const unsigned long n = data.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		fprintf(stderr, "%c", (0x80 & (data[i])) ? '1' : '0');
		fprintf(stderr, "%c", (0x40 & (data[i])) ? '1' : '0');
		fprintf(stderr, "%c", (0x20 & (data[i])) ? '1' : '0');
		fprintf(stderr, "%c", (0x10 & (data[i])) ? '1' : '0');
		fprintf(stderr, "%c", (0x08 & (data[i])) ? '1' : '0');
		fprintf(stderr, "%c", (0x04 & (data[i])) ? '1' : '0');
		fprintf(stderr, "%c", (0x02 & (data[i])) ? '1' : '0');
		fprintf(stderr, "%c", (0x01 & (data[i])) ? '1' : '0');
	}
	fprintf(stderr, "\n");
	fflush(stderr);
}
#endif

/******************************************************************************
SPA_ObfuscationEncrypter
******************************************************************************/
/******************************************************************************
SPA_ObfuscationEncrypter::SPA_ObfuscationEncrypter(void)
******************************************************************************/
SPA_ObfuscationEncrypter::SPA_ObfuscationEncrypter(void)
{
}

/******************************************************************************
SPA_ObfuscationEncrypter::~SPA_ObfuscationEncrypter(void)
******************************************************************************/
SPA_ObfuscationEncrypter::~SPA_ObfuscationEncrypter(void)
{
}

/******************************************************************************
GPL_Boolean SPA_ObfuscationEncrypter::Encrypt(const GPL_Block & key,
												GPL_Block & data)
******************************************************************************/
GPL_Boolean SPA_ObfuscationEncrypter::Encrypt(const GPL_Block & key,
												GPL_Block & data)
{
	if (0 == data.GetSize())
	{
		return FALSE;
	}
//fprintf(stderr,"#1: ");printBytes(data);

	// Pad the data.
	spaPad(data);
//fprintf(stderr,"#2: ");printBytes(data);

	// Insert a check-sum.
	const GPL_Byte checkSum = (0xFF & data.CheckSum());
	data.AppendElement(checkSum);
	data.InsertElement(checkSum);
//fprintf(stderr,"#3: ");printBytes(data);

	// Shift the data.
	unsigned long shift = randomizer.RandomNumber(1, 7);
	data.ShiftLeft(shift);
//fprintf(stderr,"#4: ");printBytes(data);

	// Insert the shift factor.
	data.InsertElement((GPL_Byte) (0xF8 & randomizer.RandomNumber()) | (0x07 & shift));
//fprintf(stderr,"#5: ");printBytes(data);
	return TRUE;
}

/******************************************************************************
GPL_Boolean SPA_ObfuscationEncrypter::Decrypt(const GPL_Block & key,
												GPL_Block & data)
******************************************************************************/
GPL_Boolean SPA_ObfuscationEncrypter::Decrypt(const GPL_Block & key,
												GPL_Block & data)
{
	if (4 > data.GetSize())
	{
		return FALSE;
	}
//fprintf(stderr,"#5: ");printBytes(data);

	// Retrieve the shift factor.
	unsigned long shift = (0x07 & data.operator[](0));
	data.RemoveElement(0);
//fprintf(stderr,"#4: ");printBytes(data);

	// Unshift the data.
	data.ShiftRight(shift);
//fprintf(stderr,"#3: ");printBytes(data);

	// Remove the check-sum.
	if (data.operator[](0) != data[data.GetSize() - 1])
	{
		return FALSE;
	}
	data.RemoveElement(data.GetSize() - 1);
	data.RemoveElement(0);
//fprintf(stderr,"#2: ");printBytes(data);

	// Unpad the data.
	spaUnpad(data);
//fprintf(stderr,"#1: ");printBytes(data);

	return TRUE;
}

GPL_OBJECT_IMPLEMENTATION(SPA_ObfuscationEncrypter, 0x40020001, "SPA_ObfuscationEncrypter");
