/******************************************************************************

Coral Tree spa/xspa.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <spa/spa.h>

static void spaTestNotary(char *s)
{
	GPL_String string = s;

	GPL_Block bytes;
	bytes.Parse(string);

	GPL_Message message;
	message.SetBuffer(bytes);

	SPA_Notary notary;
	if (!notary.Notarize(message, bytes))
	{
		fprintf(stderr, "Unable to notarize!\n");
	}
	gplByteArrayPrint(bytes, string);
	fprintf(stderr, "%s\n", string.GetData());
}

int main(void)
{
	SPA_Scalar i;
	/* VECTOR TEST */
	SPA_RowVector<3, SPA_Scalar> location;
	i = location[2];
	fprintf(stderr,"value is %f\n", i);
	location[2] = 3.45;
	i = location[2];
	fprintf(stderr,"value is %f\n", i);

	/* MATRIX TEST */
	SPA_Matrix<4,5,SPA_Scalar> matrix;
	SPA_Matrix<4,3,SPA_Scalar> answer;
	SPA_Matrix<5,3,SPA_Scalar> operand;

	matrix.Multiply(&answer, &operand);
	i = answer(2,1);
	fprintf(stderr,"value is %f\n", i);
	answer(2,0) = 2.345;
	i = answer(2,0);
	fprintf(stderr,"value is %f\n", i);
	answer(1,1) = 5.76;
	i = answer(1,1);
	fprintf(stderr,"value is %f\n", i);
	

	/* TEST OUT OF BOUNDS */
	//answer[5][1] = 1.23;
	//answer[1][5] = 3.21;

	answer(5,1) = 1.23;
	answer.Element(1,5) = 3.21;

	/* ENCRYPTION TEST */
	GPL_Message message;
	message.AppendString("SUCCESS");
	GPL_Block bytes;
	message.CopyBuffer((GPL_Array<GPL_Byte> &) bytes);
	SPA_ObfuscationEncrypter encrypter;
	if (encrypter.Encrypt(bytes))
	{
		if (encrypter.Decrypt(bytes))
		{
			message.SetBuffer((char *) bytes.GetData(), bytes.GetSize());
			const GPL_String string = message.ReadString();
			fprintf(stderr, "\"%s\"\n", string.GetData());
		}
		else
		{
			fprintf(stderr, "Unable to decrypt!\n");
		}
	}
	else
	{
		fprintf(stderr, "Unable to encrypt!\n");
	}

	/* NOTARIZATION TEST */
	spaTestNotary("00110001002E7368737472746162002E");
	spaTestNotary("74657874002E64617461002E65786365");
	spaTestNotary("7074696F6E5F72616E676573002E7374");
	spaTestNotary("61622E696E646578002E737461622E69");
	spaTestNotary("6E646578737472002E696E6974002E66");
	spaTestNotary("696E69002E636F6D6D656E74002E7374");
	spaTestNotary("72746162002E73796D746162002E7265");
	spaTestNotary("6C612E74657874002E72656C612E6461");
	spaTestNotary("7461002E72656C612E65786365707469");
	spaTestNotary("6F6E5F72616E676573002E72656C612E");
	spaTestNotary("696E6974002E72656C612E66696E6900");

	/* END TESTS */
	return 0;
}
